import mutable.v2.phase2.actions as actions #@UnresolvedImport
import mutable.v2.phase2.gui as gui #@UnresolvedImport
import mutable.v2.phase2.stzip as stzip #@UnresolvedImport

import sys
import os
import time
import traceback

VERSION = "1.0"


# Used to indicate that an error is known.
class KnownError(Exception):
    pass

# This is used to hold various information about phase 2. See what we
# stick on it in main.
class P2(object):
    pass

    
# These must be extracted early, if they exist in the package.
EXEC_FILES = [ 'usr/bin/bspatch' ]

class OutputRedirect(object):
    """
    This class looks like a file descriptor, but logs lines to the python
    log.
    """
    
    def __init__(self):
        self.stderr = sys.stderr
        self.f = open("/var/log/upgrade.log", "wb")
        
    def write(self, s):
        self.stderr.write(s)
        self.f.write(s)
        self.f.flush()

    def writelines(self, lines):
        for i in lines:
            self.write(i)
            
    def flush(self):
        return

    def close(self):
        pass

def upgrade():

    log = OutputRedirect()
    log.write("Log is open.\n")
    sys.stdout = log
    sys.stderr = log
    log.write("Log has redirected.\n")
    
    print "In phase 2."
    
    # Create the p2 object.
    p2 = P2()

    # The solartech update ZipFile storing the update.
    p2.stu = stzip.STUZipFile(sys.argv[1], "r", "mutable/v2/")

    # The root that the update will be applied to.
    p2.root = sys.argv[2]

    # A new package, if we want to be able to preserve the old state
    # of the system.
    p2.pstu = None

    # The manifest of pstu.
    p2.pmanifest = None

    # Are we an inhouse release?
    p2.inhouse = False

    # Extract the executable files needed for this upgrade.
    for fn in EXEC_FILES:
        print "Early Extracting", fn    
    
        if p2.stu.getinfo_relative("data/" + fn):
            try:
            
                f = file(p2.root + fn, "wb")
                f.write(p2.stu.read("data/" + fn))
                f.close()
            
                os.chmod(p2.root + fn, 0777)

            except:
                traceback.print_exc()
                raise
            
            print "Finished early extracting", fn

        else:
            print "Couldn't find", fn, "in stu."




    print "Initializing the gui."
    
    # Init the gui.    
    gui.init()
    gui.phase2()

    print "Basic checks."
    
    # Perform these basic checks.
    from_versions = [ i.strip() for i in p2.stu.read("meta/from").split() ]
    to_version = p2.stu.read("meta/to").strip()
    current_version = file(p2.root + "/etc/solartech/version").read().strip()

    p2.inhouse = "inhouse" in to_version

    # We shouldn't check this. In a situation where we're downgrading a
    # failed upgrade, from_version may be equal to to_version.
    
    # if current_version == to_version:
    #    ui.status("Update not needed.", color=ui.GREEN)
    #    return
    bad_fpga_fix(p2, current_version)
    
    print "Checking versions."

    if (current_version not in from_versions) and ("ANY" not in from_versions):
        gui.wrong_version()

        raise KnownError("System is at version %r, update is for %s." % 
                         (current_version, " ".join([ str(i) for i in from_versions])))

    print "Checking space."
    
    space_needed = int(p2.stu.read("meta/space_needed"))
    
    f = os.popen("df", "r")
    space_available = 100 * 1024 * 1024 * 1024

    for l in f:
        l = l.rstrip().split()

        print l
        
        if l[4] == "/":
            sa_field = l[2].replace("(DUP!)", "")
            space_available = int(sa_field) * 1024
            break
            
    f.close()

    if space_available < space_needed:
        gui.flash_full()
        raise KnownError("There is not enough free space to install this update.")

    print "Performing upgrade actions."

    actions.perform(p2)
    gui.progress(1, 1)

    print "Running upgrade script."

    # Close out the source archive.
    p2.stu.close()
    if p2.pstu is not None:
        p2.pstu.close()

    # Wait 5s for memory to stabilize.
    time.sleep(5)

    # Exec the upgrade script. And that's all, folks.
    os.execl(p2.root + "/etc/upgrade.sh", p2.root + "/etc/upgrade.sh", str(current_version), str(to_version), p2.root)

def bad_fpga_fix(p2, current_version):
    """
    This fixes a problem where the FPGA had failed enough to prevent upgrades
    from working.
    """
    
    BAD_VERSIONS = [
        "Agility-2.5.4",
        "TRAFIX-2.5.4",
        "TRAFIX-2.5.5",
        "TRAFIX-2.5.6",
        "TRAFIX-2.5.7",
        "TRAFIX-2.5.8",
        ]
    
    if current_version not in BAD_VERSIONS:
        print "FPGA fix not needed."
        return
    
    if os.path.exists("/flash/fix25"):
        print "FPGA fix performed."
        return
    
    try:
        p2.stu.getinfo_relative("data/flash/fpga.rbf")
    except:
        print "No FPGA file."
        return
    
    print "Writing FPGA fix."
    
    p2.stu.read_into("data/flash/fpga.rbf", "/tmp/fpga.rbf")
    
    print "Flashing FPGA."
    
    os.system("/bin/sync")
    os.system("/usr/bin/upgrade_flash_new -f /tmp/fpga.rbf")

    f = open("/flash/fix25", "w")
    f.close()
    
    print "Done with FPGA fix."
    
    os.system("/bin/sync")
    os.system("/bin/sync")
    os.system("/usr/bin/rtc_reboot")

    time.sleep(60)

def reboot_thread():
    time.sleep(20 * 60)
    
    
    os.system("/bin/sync")
    os.system("/bin/sync")
    os.system("/usr/bin/rtc_reboot")
    
    

def main():
    try:

        # Redirect the output to the console.
        sys.stdout.flush()
        sys.stderr.flush()
        
        f = file("/dev/console", "w", 0)
        os.dup2(f.fileno(), 1)
        os.dup2(f.fileno(), 2)

        # Do the upgrade.
        import threading
        threading.Thread(target=reboot_thread).start()
        
        upgrade()
    except SystemExit:
        pass

    except KnownError:
        gui.close()

        traceback.print_exc()

        os.system("/bin/sync")
        os.system("/bin/sync")

        os.system("/usr/bin/report_status_on_sign FAILURE")

        time.sleep(60)
        
        os.system("/usr/bin/rtc_reboot")

        # Code might get here if running report_status_on_sign fails somehow.
        while True:
            time.sleep(86400)

    except:
        gui.error()
        gui.close()
        
        traceback.print_exc()

        os.system("/bin/sync")
        os.system("/bin/sync")
        
        os.system("/usr/bin/report_status_on_sign FAILURE")

        time.sleep(60)
        
        os.system("/usr/bin/rtc_reboot")

